import threading
import types
import queue
import multiprocessing


# Only a way to handle threads inside processes (useful when can read the entire wordlist)
class ThreadHandler(threading.Thread):
    def __init__(self, wordlist_queue: multiprocessing.Queue,
                 function_: types.FunctionType, print_queue: multiprocessing.Queue):
        # Initialize the original thread elements
        threading.Thread.__init__(self)
        self.__wordlist_queue = wordlist_queue
        self.__function = function_
        self.__print_queue = print_queue

    def run(self):
        while True:
            try:
                arguments = self.__wordlist_queue.get(True, timeout=5)
            except queue.Empty:
                break
            result = self.__function(*arguments)
            if result is not None:
                self.__print_queue.put(result)
        exit(0)



import typing


class IterableWordlists(typing.Dict):
    argument: int or str
    wordlist_iterable: typing.Iterable


class BruteforceWordlists(typing.Dict):
    argument: int or str
    rule: str


class FilesWordlists(typing.Dict):
    argument: int or str
    file_path: str



import collections
import inspect
import string
import bftool
import bftool.Types
import bftool.Modes
import bftool.WordlistHandler


class Arguments(object):
    def __init__(self,
                 function_: collections.abc.Callable = None,
                 script_path: str = None,
                 function_name: str = None,
                 wordlists_iterables: bftool.Types.IterableWordlists = None,
                 wordlists_pure_bruteforce_rules: bftool.Types.BruteforceWordlists = None,
                 wordlists_files: bftool.Types.FilesWordlists = None,
                 maximum_number_of_concurrent_processes: int = 1,
                 maximum_number_of_process_threads: int = 1,
                 fuzzing_mode: bftool.Modes = bftool.Modes.ARGUMENTS_MODE,
                 ):
        if isinstance(function_, collections.abc.Callable):
            self.function = function_
        elif isinstance(script_path, str) and isinstance(function_name, str):
            self.function = bftool.import_function_from_script(script_path, function_name)
        else:
            raise ValueError("No function provided")
        function_args_spec = inspect.getfullargspec(self.function)

        wordlists_for_handler = [None] * len(function_args_spec.args)
        if wordlists_iterables:
            for key, value in wordlists_iterables.items():
                if key[0] in string.digits:
                    wordlists_for_handler[int(key)] = value
                elif isinstance(key, str):
                    wordlists_for_handler[function_args_spec.args.index(key)] = value
                else:
                    raise KeyError(f"Can't index in the function arguments key {key} of type {type(key)}")
        if wordlists_pure_bruteforce_rules:
            for key, value in wordlists_pure_bruteforce_rules.items():
                if key[0] in string.digits:
                    wordlists_for_handler[int(key)] = bftool.pure_bruteforce_rule(value)
                elif isinstance(key, str):
                    wordlists_for_handler[function_args_spec.args.index(key)] = bftool.pure_bruteforce_rule(value)
                else:
                    raise KeyError(f"Can't index in the function arguments key {key} of type {type(key)}")

        if wordlists_files:
            for key, value in wordlists_files.items():
                if key[0] in string.digits:
                    wordlists_for_handler[int(key)] = bftool.get_wordlist_from_path(value)
                elif isinstance(key, str):
                    wordlists_for_handler[function_args_spec.args.index(key)] = bftool.get_wordlist_from_path(value)
                else:
                    raise KeyError(f"Can't index in the function arguments key {key} of type {type(key)}")

        if not wordlists_for_handler:
            raise ValueError("No wordlist input provided")

        if any(value is None for value in wordlists_for_handler):
            raise IndexError(f"Invalid number of wordlists provided ({len(wordlists_for_handler)})"
                             f"for a function of {len(function_args_spec.args)} arguments")

        self.wordlist_handler = bftool.WordlistHandler.WordlistHandler()
        self.wordlist_handler.wordlists = wordlists_for_handler
        self.wordlist_handler.setup()

        self.maximum_number_of_concurrent_processes = maximum_number_of_concurrent_processes
        self.maximum_number_of_process_threads = maximum_number_of_process_threads
        self.fuzzing_mode = fuzzing_mode



import threading
import types
import multiprocessing
import queue
import bftool
import bftool.Modes
import bftool.WordlistHandler
import bftool.ThreadHandler


class ProcessHandler(multiprocessing.Process):
    def __init__(self, process_id: int, function_: types.FunctionType, wordlist_queue: multiprocessing.Queue,
                 max_threads: int, mode: int, print_queue):
        multiprocessing.Process.__init__(self)
        self.__process_id = process_id
        self.__function = function_
        self.__wordlist_queue = wordlist_queue
        self.__active_threads = 0
        self.__max_threads = max_threads
        self.__mode = mode
        self.__threads_queue = queue.Queue()
        self.__print_queue = print_queue

    def _wordlist_block_mode(self):
        for index in range(self.__max_threads):
            if self.__active_threads > self.__max_threads:
                self.__threads_queue.get().join()
                self.__active_threads -= 1
            wordlist_block_thread = bftool.ThreadHandler.ThreadHandler(self.__wordlist_queue,
                                                                       self.__function, self.__print_queue)
            self.__threads_queue.put(wordlist_block_thread)
            wordlist_block_thread.start()
            self.__active_threads += 1
        while self.__active_threads > 0:
            self.__threads_queue.get().join()
            self.__active_threads -= 1

    def __arguments_mode_function(self, *args):
        result = self.__function(*args)
        if result is not None:
            self.__print_queue.put(result)
        exit(0)

    def _arguments_block_mode(self):
        while True:
            try:
                arguments = self.__wordlist_queue.get(True, timeout=5)
            except queue.Empty:
                break
            if self.__active_threads >= self.__max_threads:
                # print(self.__active_threads)
                self.__threads_queue.get().join()
                self.__active_threads -= 1
            arguments_thread = threading.Thread(target=self.__arguments_mode_function, args=arguments)
            self.__threads_queue.put(arguments_thread)
            arguments_thread.start()
            self.__active_threads += 1
        while self.__active_threads > 0:
            self.__threads_queue.get().join()
            self.__active_threads -= 1

    def run(self):
        if self.__mode == bftool.Modes.WORDLIST_BLOCK:
            self._wordlist_block_mode()
        elif self.__mode == bftool.Modes.ARGUMENTS_MODE:
            self._arguments_block_mode()
        else:
            raise KeyError("Unknown mode supplied")
        self.__print_queue.put(f"* Process {self.__process_id} - Done")
        exit(0)



import itertools
import bftool


class WordlistHandler(object):
    def __init__(self):
        self.wordlists = []
        self.__master_wordlist = None

    def __iter__(self):
        return self.__master_wordlist

    def __next__(self):
        return next(self.__master_wordlist)

    def wordlist_from_files(self, *args):
        for wordlist_path in args:
            self.wordlists.append(bftool.read_file_lines(wordlist_path))

    def pure_bruteforce_setup(self, *args):
        for rule in args:
            self.wordlists.append(bftool.pure_bruteforce_rule(rule))

    def wordlist_from_iterable(self, *args):
        self.wordlists.extend(args)

    def setup(self):
        # self.__master_wordlist = bftool.custom_product(*self.__wordlists)  # At the end this never work
        self.__master_wordlist = itertools.product(*self.wordlists)

    def get(self):
        try:
            return next(self.__master_wordlist)
        except StopIteration:
            return None



import types
import itertools
import importlib
import os
import argparse
import copy
import multiprocessing
import bftool.Types
import bftool.Modes
import bftool.ArgumentConstructor
import bftool.WordlistHandler


def get_wordlist_from_path(wordlist_path: str):
    with open(wordlist_path, "r", errors="ignore") as file_obj:
        content = map(lambda word: word[:-1], file_obj.readlines())
        file_obj.close()
    return content


# Get the wordlists from the files supplied
def get_wordlists_from_paths(wordlists_paths: tuple) -> iter:
    for wordlist_path in wordlists_paths:
        yield get_wordlist_from_path(wordlist_path)


# Import a python module from it's filesystem path
def import_module_from_path(main_py_path: str) -> types.ModuleType:
    if not os.path.exists(main_py_path):
        raise FileExistsError("Looks like file not exists")
    if not os.path.isfile(main_py_path):
        raise TypeError("Looks like path supplied is not a file")
    module_name = main_py_path.replace("\\", "/").split("/")[-1].split(".")[0]
    return importlib.import_module(module_name, main_py_path)


def import_function_from_script(script_path: str, function_name: str):
    module = import_module_from_path(script_path)
    return getattr(module, function_name)


# Split a wordlist in smaller sub wordlists
def wordlist_divider(wordlist: tuple, step: int) -> tuple:
    length = len(wordlist)
    if step > length:
        raise IndexError("The division step can't be higher than the wordlist length")
    wordlist_step = int(length / step)
    for number in range(length):
        if wordlist_block := wordlist[number*wordlist_step:(number+1)*wordlist_step]:
            yield wordlist_block
        else:
            break


# Function to merge wordlist
def merge_wordlists(*args):
    return tuple(itertools.product(*args))


# chars, minlength, maxlength
def pure_bruteforce_rule(rule: str):
    rule = dict(value.split("=") for value in rule.split(","))
    rule["minlength"] = int(rule["minlength"])
    rule["maxlength"] = int(rule["maxlength"])
    for length in range(rule["minlength"], rule["maxlength"]+1):
        for word in itertools.product(*(rule["chars"] for number in range(length))):
            yield "".join(word)


def read_file_lines(file_path: str):
    file_object = open(file_path, errors="ignore")
    for line in file_object:
        yield line[:-1]
    file_object.close()


# By StackOverFlow user @jfs
# https://stackoverflow.com/questions/533905/get-the-cartesian-product-of-a-series-of-lists
def custom_product(*args):
    copied = (copy.copy(iterable) for iterable in args)
    if not args:
        return iter(((),))  # yield tuple()
    return (items + (item,)
            for items in custom_product(*copied[:-1]) for item in args[-1])


def arguments_queue_handler(arguments_queue: multiprocessing.Queue,
                            wordlist_handler: bftool.WordlistHandler.WordlistHandler):
    for argument in wordlist_handler:
        arguments_queue.put(argument)
    exit(0)


# Default argument capture for the main function
def get_arguments() -> bftool.ArgumentConstructor.Arguments:
    argument_parser = argparse.ArgumentParser()
    argument_parser.add_argument("-mt", "--max-threads",
                                 help="Maximum number of threads per process (if mode is set to wordlist block, \
                                 this will be also the worlist division number)", default=1, type=int)
    argument_parser.add_argument("-mp", "--max-processes",
                                 help="Maximum number of process to have active at the same time",
                                 default=bftool.Modes.ARGUMENTS_MODE, type=int)
    argument_parser.add_argument("-w", "--wordlist", help="File wordlist to use"
                                                          " based on \"[ARGUMENT_INDEX, ARGUMENT_NAME]:file_path\"",
                                 action="append", default=[])
    argument_parser.add_argument("-b", "--bruteforce",
                                 help="Generate a virtual wordlist based on \
                                 rules \"[ARGUMENT_INDEX, ARGUMENT_NAME]:chars=...,minlength=...,maxlength=...\"",
                                 action="append", default=[])
    argument_parser.add_argument("-m", "--mode",
                                 help="Mode to use during the function execution (way to divide the threads)",
                                 choices=("wordlist", "arguments"), default="arguments")
    argument_parser.add_argument("script_path", help="Path to the python script to the function to use")
    argument_parser.add_argument("function_name", help="Name of the function to use")
    parsed_arguments = argument_parser.parse_args()
    if parsed_arguments.mode == "wordlist":
        parsed_arguments.mode = bftool.Modes.WORDLIST_BLOCK
    elif parsed_arguments.mode == "arguments":
        parsed_arguments.mode = bftool.Modes.ARGUMENTS_MODE
    arguments = bftool.ArgumentConstructor.Arguments(script_path=parsed_arguments.script_path,
                                                     function_name=parsed_arguments.function_name,
                                                     wordlists_files=bftool.Types.FilesWordlists(wordlist.split(":")
                                                                          for wordlist in parsed_arguments.wordlist),
                                                     wordlists_pure_bruteforce_rules=bftool.Types.BruteforceWordlists(
                                                         wordlist.split(":")
                                                         for wordlist in parsed_arguments.bruteforce),
                                                     maximum_number_of_concurrent_processes=parsed_arguments.max_processes,
                                                     maximum_number_of_process_threads=parsed_arguments.max_threads,
                                                     fuzzing_mode=parsed_arguments.mode,
                                                     )
    return arguments



from .useful_functions import wordlist_divider
from .useful_functions import merge_wordlists
from .useful_functions import import_module_from_path
from .useful_functions import import_function_from_script
from .useful_functions import get_wordlist_from_path
from .useful_functions import get_wordlists_from_paths
from .useful_functions import get_arguments
from .useful_functions import read_file_lines
from .useful_functions import pure_bruteforce_rule
from .useful_functions import custom_product
from .useful_functions import arguments_queue_handler



import multiprocessing
import time
import threading
import sys
import bftool
import bftool.ArgumentConstructor
import bftool.WordlistHandler
import bftool.ProcessHandler


class MainHandler(object):
    def __init__(self):
        # Wordlist distributed processes
        self.__processes = []
        self.__print_queue = multiprocessing.Queue()
        self.__finish = False

    # Use this in win32 system as it has trouble using multiprocessing.Process().start()
    def start_win32_process(self, process: multiprocessing.Process):
        process_thread = threading.Thread(target=process.run, )
        self.__processes.append(process_thread)

    # Use this function to handle prints
    def print_queue_handler(self):
        while not self.__finish:
            print(self.__print_queue.get())
        while not self.__print_queue.empty():
            print(self.__print_queue.get())

    def main(self, arguments: bftool.ArgumentConstructor.Arguments = None):
        total_time = time.time()
        if arguments is None:
            arguments = bftool.get_arguments()

        # Get the function to be used by the handlers
        function_ = arguments.function

        wordlist_queue = multiprocessing.Queue()

        # Start filling the queue
        wordlist_queue_thread = threading.Thread(target=bftool.arguments_queue_handler,
                                                 args=(wordlist_queue, arguments.wordlist_handler)
                                                 )
        wordlist_queue_thread.start()

        self.__print_queue.put("--- Starting child processes ---")
        print_queue_thread = threading.Thread(target=self.print_queue_handler, )
        print_queue_thread.start()
        processes_setup_time = time.time()
        for index in range(1, arguments.maximum_number_of_concurrent_processes+1):
            process_handler = bftool.ProcessHandler.ProcessHandler(index, function_, wordlist_queue,
                                                                   arguments.maximum_number_of_process_threads,
                                                                   arguments.fuzzing_mode,
                                                                   self.__print_queue)
            if sys.platform == "win32":
                self.start_win32_process(process_handler)
            else:
                self.__processes.append(process_handler)

            self.__print_queue.put(f"* Process with ID {index} - Prepared")
        processes_setup_time = time.time() - processes_setup_time
        self.__print_queue.put("--- All processes were prepared ---")
        self.__print_queue.put("--- Waiting to finish ---")
        fuzzing_time = time.time()
        for process in self.__processes:
            process.start()
        for process_handler in self.__processes:
            process_handler.join()
        fuzzing_time = time.time() - fuzzing_time
        wordlist_queue_thread.join()
        total_time = time.time() - total_time
        self.__print_queue.put("--- END ---")
        self.__print_queue.put(f"Time setting up the processes: {processes_setup_time}")
        self.__print_queue.put(f"Time fuzzing: {fuzzing_time}")
        self.__print_queue.put(f"Total time: {total_time}")
        self.__finish = True
        print_queue_thread.join()
        exit(0)
